import pytest

from mock import patch, Mock

from endpoints.api.test.shared import client_with_identity, conduct_api_call
from endpoints.api.tag import RepositoryTag, RestoreTag
from features import FeatureNameValue

from test.fixtures import *


@pytest.fixture()
def get_repo_image():
  def mock_callable(namespace, repository, image_id):
    img = Mock(repository='fetched_repository') if image_id == 'image1' else None
    return img

  with patch('endpoints.api.tag.model.image.get_repo_image', side_effect=mock_callable) as mk:
    yield mk


@pytest.fixture()
def get_repository():
  with patch('endpoints.api.tag.model.image.get_repo_image', return_value='mock_repo') as mk:
    yield mk


@pytest.fixture()
def get_repo_tag_image():
  def mock_get_repo_tag_image(repository, tag):
    tag_img = Mock(docker_image_id='mock_docker_image_id') if tag == 'existing-tag' else None
    return tag_img

  with patch('endpoints.api.tag.model.tag.get_repo_tag_image',
             side_effect=mock_get_repo_tag_image):
    yield


@pytest.fixture()
def restore_tag_to_manifest():
  def mock_restore_tag_to_manifest(repository, tag, manifest_digest):
    tag_img = Mock(docker_image_id='mock_docker_image_id') if tag == 'existing-tag' else None
    return tag_img

  with patch('endpoints.api.tag.model.tag.restore_tag_to_manifest',
             side_effect=mock_restore_tag_to_manifest):
    yield


@pytest.fixture()
def restore_tag_to_image():
  def mock_restore_tag_to_image(repository, tag, image_id):
    tag_img = Mock(docker_image_id='mock_docker_image_id') if tag == 'existing-tag' else None
    return tag_img

  with patch('endpoints.api.tag.model.tag.restore_tag_to_image',
             side_effect=mock_restore_tag_to_image):
    yield


@pytest.fixture()
def create_or_update_tag():
  with patch('endpoints.api.tag.model.tag.create_or_update_tag') as mk:
    yield mk


@pytest.fixture()
def generate_manifest():
  def mock_callable(namespace, repository, tag):
    if tag == 'generatemanifestfail':
      raise Exception('test_failure')

  with patch('endpoints.api.tag._generate_and_store_manifest', side_effect=mock_callable) as mk:
    yield mk


@pytest.fixture()
def authd_client(client):
  with client_with_identity('devtable', client) as cl:
    yield cl


@pytest.mark.parametrize('test_image,test_tag,expected_status', [
  ('image1', '-INVALID-TAG-NAME', 400),
  ('image1', '.INVALID-TAG-NAME', 400),
  ('image1',
   'INVALID-TAG_NAME-BECAUSE-THIS-IS-WAY-WAY-TOO-LOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOONG',
   400),
  ('nonexistantimage', 'newtag', 404),
  ('image1', 'generatemanifestfail', None),
  ('image1', 'existing-tag', 201),
  ('image1', 'newtag', 201),
])
def test_move_tag(test_image, test_tag, expected_status, get_repo_image, get_repo_tag_image,
                  create_or_update_tag, generate_manifest, authd_client):
  params = {'repository': 'devtable/repo', 'tag': test_tag}
  request_body = {'image': test_image}
  if expected_status is None:
    with pytest.raises(Exception):
      conduct_api_call(authd_client, RepositoryTag, 'put', params, request_body, expected_status)
  else:
    conduct_api_call(authd_client, RepositoryTag, 'put', params, request_body, expected_status)


@pytest.mark.parametrize('test_manifest,test_tag,manifest_generated,expected_status', [
  (None, 'newtag', True, 200),
  (None, 'generatemanifestfail', True, None),
  ('manifest1', 'newtag', False, 200),
])
def test_restore_tag(test_manifest, test_tag, manifest_generated, expected_status, get_repository,
                     restore_tag_to_manifest, restore_tag_to_image, generate_manifest,
                     authd_client):
  params = {'repository': 'devtable/repo', 'tag': test_tag}
  request_body = {'image': 'image1'}
  if test_manifest is not None:
    request_body['manifest_digest'] = test_manifest
  if expected_status is None:
    with pytest.raises(Exception):
      conduct_api_call(authd_client, RestoreTag, 'post', params, request_body, expected_status)
  else:
    conduct_api_call(authd_client, RestoreTag, 'post', params, request_body, expected_status)

  if manifest_generated:
    generate_manifest.assert_called_with('devtable', 'repo', test_tag)
